
/* Program to compute H(codon_n|codon_a,prev_codon_n) -- Peter N. Yianilos 1996.
   Part of a joint project with David Loewenstern. */

#include <stdio.h>
#include <assert.h>
#include <math.h>

#include "util.h"
#include "amino.h"

char table[64][5] = {
	"aaaK", "aacN", "aagK", "aatN", "acaT", "accT", "acgT", "actT",
	"agaR", "agcS", "aggR", "agtS", "ataI", "atcI", "atgM", "attI",
	"caaQ", "cacH", "cagQ", "catH", "ccaP", "cccP", "ccgP", "cctP",
	"cgaR", "cgcR", "cggR", "cgtR", "ctaL", "ctcL", "ctgL", "cttL",
	"gaaE", "gacD", "gagE", "gatD", "gcaA", "gccA", "gcgA", "gctA",
	"ggaG", "ggcG", "gggG", "ggtG", "gtaV", "gtcV", "gtgV", "gttV",
	"taa-", "tacY", "tag-", "tatY", "tcaS", "tccS", "tcgS", "tctS", 
	"tga-", "tgcC", "tggW", "tgtC", "ttaL", "ttcF", "ttgL", "tttF",
};

#define NAMINO  20
#define NCODONS 64

void usage(char *prog)
{
	fprintf(stderr, "Usage is: %s <dnafile> <#cross-segments> <context-mask>\n", prog);
	exit(-1);
}

int scancodon(char *output, FILE *in)
{
	int i, r;

	for (i = 0; i < 3; ++i) {
		while (1) {
			if ((r = fgetc(in)) == EOF)
			  return(0);
			if (!isspace(r))
			  break;
		}
		switch(r) {
		      case 'a':
			r = 0;
			break;
		      case 'c':
			r = 1;
			break;
		      case 'g':
			r = 2;
			break;
		      case 't':
			r = 3;
			break;
		      default:
			error_fatal("scancodon", "Bad base in input file: %c\n", r);
		}
		output[i] = r;
	}
	return(1);
}

int aminovalue(char amino)
{
	switch(amino) {
	      case 'A': return(0);
	      case 'C': return(1);
	      case 'D': return(2);
	      case 'E': return(3);
	      case 'F': return(4);
	      case 'G': return(5);
	      case 'H': return(6);
	      case 'I': return(7);
	      case 'K': return(8);
	      case 'L': return(9);
	      case 'M': return(10);
	      case 'N': return(11);
	      case 'P': return(12);
	      case 'Q': return(13);
	      case 'R': return(14);
	      case 'S': return(15);
	      case 'T': return(16);
	      case 'V': return(17);
	      case 'W': return(18);
	      case 'Y': return(19);
	      default: error_fatal("aminovalue", "Bad amino acid character %c\n", amino);
	}
}

int codonindex(char *codon)
{
	return(codon[0]*16 + codon[1]*4 + codon[2]);
}

int basecontexts(int mask)
{
	switch (mask) {
	      case 0: return 1;
	      case 1: return 4;
	      case 2: return 4;
	      case 3: return 16;
	      case 4: return 4;
	      case 5: return 16;
	      case 6: return 16;
	      case 7: return 64;
	      default: assert (1 == 2);
	}
}

int contextindex(char *codon, int mask)
{
	switch (mask) {
	      case 0: return 0;
	      case 1: return codon[2];
	      case 2: return codon[1];
	      case 3: return codon[1]*4 + codon[2];
	      case 4: return codon[0];
	      case 5: return codon[0]*4 + codon[2];
	      case 6: return codon[0]*4 + codon[1];
	      case 7: return codon[0]*16 + codon[1]*4 + codon[2];
	      default: assert (1 == 2);
	}
}

int aminoindex(char *codon)
{
	return aminovalue(table[codonindex(codon)][3]);
}

struct codon {
	char c[3];
};

int main(int argn, char *argv[])
{
	int cross_segments, context_mask, segment;
	int i, j, k, test_start, test_end, total_codons;
	int bvals;
	double ***model, total, grand;
	int x1, x2, x3;
	struct codon *dna, a_codon, *train, *test;
	FILE *dnafile;

	/* Process arguments */

	if (argn != 4)
	  usage(argv[0]);

	cross_segments = atoi(argv[2]);

	if (cross_segments < 2) {
		fprintf(stderr, "Must specify at least 2 cross validation segments\n");
		usage(argv[0]);
	}

	context_mask = atoi(argv[3]);

	if (context_mask < 0 || context_mask > 7) {
		fprintf(stderr, "Must specify context-mask in range 0-7\n");
		usage(argv[0]);
	}
	
	if ((dnafile = fopen(argv[1], "r")) == NULL) {
		fprintf(stderr, "Can't open dna file %s\n", argv[1]);
		usage(argv[0]);
	}

	/* Read in the DNA */

	dna = vector_create(sizeof(struct codon), 0, 0);

	while (scancodon(a_codon.c, dnafile)) 
	  dna = vector_extend(dna, &a_codon);

	total_codons = vector_length(dna);
	printf("%d codons read\n", total_codons);
	
	/* Do cross-validation */

	for (segment = 0, grand = 0; segment < cross_segments; ++segment) {
		
		/* Establish range for withheld test segment */
		
		test_start = (total_codons/cross_segments) * segment;
		test_end   = (total_codons/cross_segments) * (segment+1);
		
		if (test_end >= total_codons)
		  test_end = total_codons - 1;
		
		/* Build training and test segments */

		train = vector_create(sizeof(struct codon), 0, 0);
		test  = vector_create(sizeof(struct codon), 0, 0);

		for (i = 0; i < total_codons; ++i) {
			if (i >= test_start && i <= test_end)
			  test  = vector_extend(test, dna[i].c);
			else
			  train = vector_extend(train, dna[i].c);
		}
		
		printf("Leg %2d/%2d -- %6d train %6d test: ", segment, cross_segments, 
		       vector_length(train), vector_length(test));
		
		/* Allocate the model */

		bvals = basecontexts(context_mask);
		
		model = vector_create(sizeof(double **), 0, bvals);
		for (i = 0; i < bvals; ++i) {
			model[i] = vector_create(sizeof(double *), 0, NAMINO);
			for (j = 0; j < NAMINO; ++j) {
				model[i][j] = vector_create(sizeof(double), 0, NCODONS);
				for (k = 0; k < NCODONS; ++k) {
					if (isupper(table[k][3]))
					  model[i][j][k] = 1E-6;
					else
					  model[i][j][k] = 0;
				}
			}
		}

		/* Learn the model */

		for (i = 0, total = 0; i < vector_length(train)-1; ++i) {
			
			x1 = contextindex(train[i].c, context_mask);
			assert(x1 >= 0 && x1 < vector_length(model));

			x2 = aminoindex(train[i+1].c);
			assert(x2 >= 0 && x2 < vector_length(model[x1]));
			
			x3 = codonindex(train[i+1].c);
			assert(x3 >= 0 && x3 < vector_length(model[x1][x2]));
			
			assert(model[x1][x2][x3] != 0);
			model[x1][x2][x3] += 1;
		}

		/* Normalize the predictions (and leave bits in the array) */

		for (i = 0; i < bvals; ++i) {
			for (j = 0; j < NAMINO; ++j) {
				for (k = 0, total = 0; k < NCODONS; ++k) 
				  total += model[i][j][k];
				for (k = 0; k < NCODONS; ++k) {
					if (model[i][j][k] != 0)
					  model[i][j][k] = -log(model[i][j][k]/total)/log(2.0);
					else
					  model[i][j][k] = -1;
				}
			}
		}
		
		/* Predict the test set */

		for (i = 0, total = 0; i < vector_length(test)-1; ++i) {
			x1 = contextindex(test[i].c, context_mask);
			assert(x1 >= 0 && x1 < vector_length(model));

			x2 = aminoindex(test[i+1].c);
			assert(x2 >= 0 && x2 < vector_length(model[x1]));
			
			x3 = codonindex(test[i+1].c);
			assert(x3 >= 0 && x3 < vector_length(model[x1][x2]));
			
			assert(model[x1][x2][x3] >= 0);
			total += model[x1][x2][x3];
		}
		
		total /= vector_length(test) - 1;

		printf("%.5f bits per codon.\n", total);
		
		grand += total;

		/* Cleanup after each cross validation leg */

		for (i = 0; i < bvals; ++i) {
			for (j = 0; j < NAMINO; ++j) {
				vector_destroy(model[i][j]);
			}
			vector_destroy(model[i]);
		}
		vector_destroy(model);
		
		vector_destroy(train);
		vector_destroy(test);
	}

	grand /= cross_segments;
	printf("Average = %.5f (%.5f bits per base)\n", grand, grand/3);

	/* Clean up */
	
	vector_destroy(dna);
	fclose(dnafile);
	assert(mem_total_space() == 0);
}

