
/* Code for computation of "amino hamming distance" - Peter N. Yianilos 1996. 
   Part of a joint project with David Loewenstern. */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <limits.h>
#include <ctype.h>

#include "util.h"
#include "amino.h"

/* The routines in this module build a very efficient and compact finite state machine to
   compute "amino hamming distance" (ahd) as we define it.  Defining such a distance
   between strings whose length is a multiple of 3, and which are assumed to be
   aligned with the reading frame, is straightforward: it is simply the number of
   mismatches in the amino acid sequence coded by the nucleotide sequence; where
   the "stop" signal is considered to be the 21st amino acid.
   
   These routines extend this definition to the case in which the strings being
   compared are not necessarily a multiple of 3 in length, and drops the assumption
   that they are aligned with the reading frame.  The comparison does however assume
   that the strings are similarly aligned.  
   
   The automaton is fed pairs of base pairs, and accumulates three hamming distances
   corresponding to the three possible alignments.
   
   So the automaton deals with two problems: 1) the fact that we may see only part of
   a codon and nevertheless have to compute a hamming distance contribution, 2) the
   fact that we must deal with three possible alignments with the reading frame.
    
   We begin by defining a finite state automaton which has as input a base pair sequence
   and after each base pair is processed, reports the possible identities of the
   amino acid corresponding to the most recently processed base pair.  Initial
   state 0 corresponds to starting at the beginning of a reading frame.  State 1
   and two correspond to late starts.  The automaton is based on the following
   table extracted by Loewenstern from published references:

   codon   amino   1-letter
   regexp  acid    abbreviation
   ------------------
   aa[ag]  Lys     K
   aa[ct]  Asn     N
   ac?     Thr     T
   ag[ag]  Arg     R
   ag[ct]  Ser     S
   at[act] Ile     I
   atg     Met     M
   ca[ag]  Gln     Q
   ca[ct]  His     H
   cc?     Pro     P
   cg?     Arg     R
   ct?     Leu     L
   ga[ag]  Glu     E
   ga[ct]  Asp     D
   gc?     Ala     A
   gg?     Gly     G
   gt?     Val     V
   ta[ag]  STOP    -
   ta[ct]  Tyr     Y
   tc?     Ser     S
   tg[ct]  Cys     C
   tga     STOP    -
   tgg     Trp     W
   tt[ag]  Leu     L
   tt[ct]  Phe     F

*/

#define A     0x00000001
#define C     0x00000002
#define D     0x00000004
#define E     0x00000008
#define F     0x00000010
#define G     0x00000020
#define H     0x00000040
#define I     0x00000080
#define K     0x00000100
#define L     0x00000200
#define M     0x00000400
#define N     0x00000800
#define P     0x00001000
#define Q     0x00002000
#define R     0x00004000
#define S     0x00008000
#define T     0x00010000
#define V     0x00020000
#define W     0x00040000
#define Y     0x00080000
#define STOP  0x10000000
#define ALL   0x100FFFFF

/* These routines implement an "amino decoder".  The three start states are 0,1,2
and correspond to starting in perfect alignment with the reading frame, and to the
two possible degrees of misalignment.  */

typedef struct {
	int state;
	int mask;
} AMINO_DECODER;

static void init_amino_decoder(AMINO_DECODER *a, int start_pos)
{
	a->state = start_pos;
	a->mask = ALL;
}

#define AMINO_DECODER_STATES 27  /* Number of states in basic amino automaton below */

static void advance_amino_decoder(AMINO_DECODER *a, int bp)
{
	assert(bp >= 0 && bp <= 3);

	switch(a->state) {
	      case 0:  /* Start state at beginning of reading frame */
		switch(bp) {
		      case 0:
			a->state = 3;
			a->mask = K+N+T+R+S+I+M;
			return;
		      case 1:
			a->state = 4;
			a->mask = Q+H+P+R+L;
			return;
		      case 2:
			a->state = 5;
			a->mask = E+D+A+G+V;
			return;
		      case 3:
			a->state = 6;
			a->mask = STOP+Y+S+C+W+L+F;
			return;
		}
	      case 1:  /* Start state displaced one into reading frame */
		switch(bp) {
		      case 0:
			a->state = 23;
			a->mask = K+N+Q+H+E+D+STOP+Y;
			return;
		      case 1:
			a->state = 24;
			a->mask = T+P+A+S;
			return;
		      case 2:
			a->state = 25;
			a->mask = R+S+G+C+STOP+W;
			return;
		      case 3:
			a->state = 26;
			a->mask = I+M+L+V+F;
			return;
		}
	      case 2:  /* Start state displaced two into reading frame */
		a->state = 0; 
		switch(bp) {
		      case 0:
			a->mask = K+T+R+I+Q+P+L+E+A+G+V+STOP+S;
			return;
		      case 1:
			a->mask = N+T+S+I+H+P+R+L+D+A+G+V+Y+C+F;
			return;
		      case 2:
			a->mask = K+T+R+M+Q+P+L+E+A+G+V+STOP+S+W;
			return;
		      case 3:
			a->mask = N+T+S+I+H+P+R+L+D+A+G+V+Y+C+F;
			return;
		}
	      case 3:  /* Have seen "a" from aligned start */
		switch(bp) {
		      case 0:
			a->state = 7;
			a->mask = K+N;
			return;
		      case 1:
			a->state = 8;
			a->mask = T;
			return;
		      case 2:
			a->state = 9;
			a->mask = R+S;
			return;
		      case 3:
			a->state = 10;
			a->mask = I+M;
			return;
		}
	      case 4:  /* Have seen "c" from aligned start */
		switch(bp) {
		      case 0:
			a->state = 11;
			a->mask = Q+H;
			return;
		      case 1:
			a->state = 12;
			a->mask = P;
			return;
		      case 2:
			a->state = 13;
			a->mask = R;
			return;
		      case 3:
			a->state = 14;
			a->mask = L;
			return;
		}
	      case 5:  /* Have seen "g" from aligned start */
		switch(bp) {
		      case 0:
			a->state = 15;
			a->mask = E+D;
			return;
		      case 1:
			a->state = 16;
			a->mask = A;
			return;
		      case 2:
			a->state = 17;
			a->mask = G;
			return;
		      case 3:
			a->state = 18;
			a->mask = V;
			return;
		}
	      case 6:  /* Have seen "t" from aligned start */
		switch(bp) {
		      case 0:
			a->state = 19;
			a->mask = STOP+Y;
			return;
		      case 1:
			a->state = 20;
			a->mask = S;
			return;
		      case 2:
			a->state = 21;
			a->mask = C+STOP+W;
			return;
		      case 3:
			a->state = 22;
			a->mask = L+F;
			return;
		}
	      case 7:  /* Have seen "aa" from aligned start */
		a->state = 0;
		switch(bp) {
		      case 0:
			a->mask = K;
			return;
		      case 1:
			a->mask = N;
			return;
		      case 2:
			a->mask = K;
			return;
		      case 3:
			a->mask = N;
			return;
		}
	      case 8:  /* Have seen "ac" from aligned start */
		a->state = 0;
		a->mask = T;
		return;
	      case 9:  /* Have seen "ag" from aligned start */
		a->state = 0;
		switch(bp) {
		      case 0:
			a->mask = R;
			return;
		      case 1:
			a->mask = S;
			return;
		      case 2:
			a->mask = R;
			return;
		      case 3:
			a->mask = S;
			return;
		}
	      case 10:  /* Have seen "at" from aligned start */
		a->state = 0;
		switch(bp) {
		      case 0:
			a->mask = I;
			return;
		      case 1:
			a->mask = I;
			return;
		      case 2:
			a->mask = M;
			return;
		      case 3:
			a->mask = I;
			return;
		}
	      case 11:  /* Have seen "ca" from aligned start */
		a->state = 0;
		switch(bp) {
		      case 0:
			a->mask = Q;
			return;
		      case 1:
			a->mask = H;
			return;
		      case 2:
			a->mask = Q;
			return;
		      case 3:
			a->mask = H;
			return;
		}
	      case 12:  /* Have seen "cc" from aligned start */
		a->state = 0;
		a->mask = P;
		return;
	      case 13:  /* Have seen "cg" from aligned start */
		a->state = 0;
		a->mask = R;
		return;
	      case 14:  /* Have seen "ct" from aligned start */
		a->state = 0;
		a->mask = L;
		return;
	      case 15:  /* Have seen "ga" from aligned start */
		a->state = 0;
		switch(bp) {
		      case 0:
			a->mask = E;
			return;
		      case 1:
			a->mask = D;
			return;
		      case 2:
			a->mask = E;
			return;
		      case 3:
			a->mask = D;
			return;
		}
	      case 16:  /* Have seen "gc" from aligned start */
		a->state = 0;
		a->mask = A;
		return;
	      case 17:  /* Have seen "gg" from aligned start */
		a->state = 0;
		a->mask = G;
		return;
	      case 18:  /* Have seen "gt" from aligned start */
		a->state = 0;
		a->mask = V;
		return;
	      case 19:  /* Have seen "ta" from aligned start */
		a->state = 0;
		switch(bp) {
		      case 0:
			a->mask = STOP;
			return;
		      case 1:
			a->mask = Y;
			return;
		      case 2:
			a->mask = STOP;
			return;
		      case 3:
			a->mask = Y;
			return;
		}
	      case 20:  /* Have seen "tc" from aligned start */
		a->state = 0;
		a->mask = S;
		return;
	      case 21:  /* Have seen "tg" from aligned start */
		a->state = 0;
		switch(bp) {
		      case 0:
			a->mask = STOP;
			return;
		      case 1:
			a->mask = C;
			return;
		      case 2:
			a->mask = W;
			return;
		      case 3:
			a->mask = C;
			return;
		}
	      case 22:  /* Have seen "tt" from aligned start */
		a->state = 0;
		switch(bp) {
		      case 0:
			a->mask = L;
			return;
		      case 1:
			a->mask = F;
			return;
		      case 2:
			a->mask = L;
			return;
		      case 3:
			a->mask = F;
			return;
		}
	      case 23:  /* Have seen "?a" from +1 start */
		a->state = 0;
		switch(bp) {
		      case 0:
			a->mask = K+Q+E+STOP;
			return;
		      case 1:
			a->mask = N+H+D+Y;
			return;
		      case 2:
			a->mask = K+Q+E+STOP;
			return;
		      case 3:
			a->mask = N+H+D+Y;
			return;
		}
	      case 24:  /* Have seen "?c" from +1 start */
		a->state = 0;
		a->mask = T+P+A+S;
		return;
	      case 25:  /* Have seen "?g" from +1 start */
		a->state = 0;
		switch(bp) {
		      case 0:
			a->mask = R+G+STOP;
			return;
		      case 1:
			a->mask = S+R+G+C;
			return;
		      case 2:
			a->mask = R+G+W;
			return;
		      case 3:
			a->mask = S+R+G+C;
			return;
		}
	      case 26:  /* Have seen "?t" from +1 start */
		a->state = 0;
		switch(bp) {
		      case 0:
			a->mask = I+L+V;
			return;
		      case 1:
			a->mask = I+L+V+F;
			return;
		      case 2:
			a->mask = M+L+V;
			return;
		      case 3:
			a->mask = I+L+V+F;
			return;
		}
	}
}

/* These routines build a "combined decoder" which runs three "amino decoders" at the
same time and collapses the result into a single machine */

typedef struct {
	int m[3];
	int s[3];
	int next_state;
} COMBINED_ENTRY;

static void
combined_fill_in(int state, int substates[3],
		 int map[AMINO_DECODER_STATES][AMINO_DECODER_STATES][AMINO_DECODER_STATES],
		 COMBINED_ENTRY ***fsm)
{
	int i, bp, ns, rss[3];
	AMINO_DECODER ad[3];
	COMBINED_ENTRY *cep = NULL;

	assert(state < vector_length(*fsm));

	(*fsm)[state] = vector_create(sizeof(COMBINED_ENTRY), 0, 4);

	map[substates[0]][substates[1]][substates[2]] = state;
	
	for (bp = 0; bp < 4; ++bp) {
		
		/* Initialize and advance all three decoders filling in masks and substates */

		for (i = 0; i < 3; ++i) {
			init_amino_decoder(&ad[i], substates[i]);
			advance_amino_decoder(&ad[i], bp);
			(*fsm)[state][bp].m[i] = ad[i].mask;
			rss[i] = ad[i].state;
			(*fsm)[state][bp].s[i] = ad[i].state;
		}

		/* Fill in state field */

		ns = map[rss[0]][rss[1]][rss[2]];

		if ( ns >= 0) { /* State already exists */

			(*fsm)[state][bp].next_state = ns;

		} else { /* Create new state and recurse */
			
			*fsm = vector_extend(*fsm, &cep);
			ns = vector_length(*fsm) - 1;
			(*fsm)[state][bp].next_state = ns;

			combined_fill_in(ns, rss, map, fsm);
		} 
	}
}

static COMBINED_ENTRY **build_combined_decoder(void)
{
	int i, j, k;
	int map[AMINO_DECODER_STATES][AMINO_DECODER_STATES][AMINO_DECODER_STATES];
	int substates[3];
	COMBINED_ENTRY **fsm;

	/* Initialize the map array which will correspond state triples to single states
	   in the combined machine */

	for (i = 0; i < AMINO_DECODER_STATES; ++i) 
	  for (j = 0; j < AMINO_DECODER_STATES; ++j) 
	    for (k = 0; k < AMINO_DECODER_STATES; ++k) 
	      map[i][j][k] = -1;

	/* Allocate an initially empty FSM */
	
	fsm = vector_create(sizeof(COMBINED_ENTRY *), 0, 1);
	
	/* Recursive call to fill in the table */

	for (i = 0; i < 3; ++i) 
	  substates[i] = i;

	combined_fill_in(0, substates, map, &fsm);

	/* Return the resulting machine */

	return(fsm);
}

static void destroy_combined_decoder(COMBINED_ENTRY **fsm)
{
	int i;

	for (i = 0; i < vector_length(fsm); ++i) 
	  vector_destroy(fsm[i]);

	vector_destroy(fsm);
}

static void print_combined_decoder(COMBINED_ENTRY **fsm)
{
	int i, j, bp;

	printf("%d states\n", vector_length(fsm));

	for (i = 0; i < vector_length(fsm); ++i) {
		printf("State %d\n", i);
		for (bp = 0; bp < 4; ++bp) {
			printf("\t");
			for (j = 0; j < 3; ++j) {
				printf("(%08x %3d) ", fsm[i][bp].m[j], fsm[i][bp].s[j]);
			}
			printf("%d\n", fsm[i][bp].next_state);
		}
	}
}

/* These routines build a "master decoder" which combines two "combined decoders" to
form a single machine which takes as input a pair of base pairs, represented as an
integer from 0 to 15.  Edges of this machine contain all of the information from
the combined decoder.  This machine is then simplified to form the "final decoder"
in routines to follow.  This simplification amounts to discarding unneeded information
and storing the three hamming distance increment signals in a single integer */

typedef struct {
	int cs[2];
	int m[2][3];
	int s[2][3];
	int hamming[3];
	int next_state;
} MASTER_ENTRY;

static void master_fill_in(int state, int ss[2], int **map,
		    MASTER_ENTRY ***mfsm, COMBINED_ENTRY **cfsm, int *hamming)
{
	int i, j, k, bp[2], b, ns, rss[2], rh[3];
	MASTER_ENTRY *mep = NULL;

	assert(state < vector_length(*mfsm));

	(*mfsm)[state] = vector_create(sizeof(MASTER_ENTRY), 0, 16);

	map[ss[0]][ss[1]] = state;
	
	for (bp[0] = 0; bp[0] < 4; ++bp[0]) {
		for (bp[1] = 0; bp[1] < 4; ++bp[1]) {

			b = bp[0]*4 + bp[1];

			/* Initialize and advance cfsms filling in masks and substates */
			
			for (i = 0; i < 2; ++i) {
				
				(*mfsm)[state][b].cs[i] = cfsm[ss[i]][bp[i]].next_state;

				rss[i] = (*mfsm)[state][b].cs[i];

				for (j = 0; j < 3; ++j) {
					(*mfsm)[state][b].m[i][j] = cfsm[ss[i]][bp[i]].m[j];
					(*mfsm)[state][b].s[i][j] = cfsm[ss[i]][bp[i]].s[j];
				}
			}
			
			/* Fill in hamming distance increments */

			for (j = 0; j < 3; ++j) {
				rh[j] = hamming[j];
				(*mfsm)[state][b].hamming[j] = 0;

				if (rh[j] == 0) { /* No difference seen yet, so look */
					if (((*mfsm)[state][b].m[0][j] &
					     (*mfsm)[state][b].m[1][j]) == 0) { /* A diff */

						rh[j] = 1;
						(*mfsm)[state][b].hamming[j] = 1;
					}
				}

				if ((*mfsm)[state][b].s[0][j] == 0) { /* Ending codon */
					assert((*mfsm)[state][b].s[1][j] == 0);
					rh[j] = 0;

				}
			}

			/* Fill in state field */
			
			ns = map[rss[0]][rss[1]];
			
			if ( ns >= 0) { /* State already exists */
				
				(*mfsm)[state][b].next_state = ns;
				
			} else { /* Create new state and recurse */

				*mfsm = vector_extend(*mfsm, &mep);
				ns = vector_length(*mfsm) - 1;
				(*mfsm)[state][b].next_state = ns;
				
				master_fill_in(ns, rss, map, mfsm, cfsm, rh);
			} 
		}
	}
}

static MASTER_ENTRY **build_master_decoder(COMBINED_ENTRY **cfsm)
{
	int i, j;
	int fs, **map;
	int substates[2], hamming[3];
	MASTER_ENTRY **mfsm;

	/* Initialize the map array which will correspond state pairs to single states
	   in the master machine */

	fs = vector_length(cfsm);
	map = vector_create(sizeof(int *), 0, fs);
	for (i = 0; i < fs; ++i)
	  map[i] = vector_create(sizeof(int), 0, fs);

	for (i = 0; i < fs; ++i) 
	  for (j = 0; j < fs; ++j) 
	    map[i][j] = -1;

	/* Allocate an initially empty FSM */
	
	mfsm = vector_create(sizeof(MASTER_ENTRY *), 0, 1);
	
	/* Recursive call to fill in the table */

	for (i = 0; i < 2; ++i) 
	  substates[i] = 0;

	for (i = 0; i < 3; ++i) 
	  hamming[i] = 0;

	master_fill_in(0, substates, map, &mfsm, cfsm, hamming);

	/* Destroy the map array */

	for (i = 0; i < fs; ++i)
	  vector_destroy(map[i]);
	vector_destroy(map);

	/* Return the resulting machine */

	return(mfsm);
}

static void destroy_master_decoder(MASTER_ENTRY **fsm)
{
	int i;

	for (i = 0; i < vector_length(fsm); ++i) 
	  vector_destroy(fsm[i]);

	vector_destroy(fsm);
}

/* These routines build our "final decoder".  By this we mean that the decoder built
by this machine is the one which applications will use.  It is designed to minimize
storage space (important for cache locality), and all three hamming distance increments
are combined into a single 32-bit int so that the user program performs a single
addition to update all three hamming distances */

/* Note the macros below pack three hamming distances into a single 32-bit int.  Each
   has 10 bits so distances of up to 1023 are possible.  Nucleotide sequences of three
   times this value are therefore supported since codon length is 3.  Actually 3*1023-2
   are supported because of the effect of misaligned codons. */

FINAL_ENTRY **build_final_decoder(void)
{
	int state, jbp, h[3], k;
	COMBINED_ENTRY **fsm;
	MASTER_ENTRY **mfsm;
	FINAL_ENTRY **ffsm;
	
	fsm = build_combined_decoder();
	mfsm = build_master_decoder(fsm);
	destroy_combined_decoder(fsm);

	ffsm = vector_create(sizeof(FINAL_ENTRY *), 0, vector_length(mfsm));

	for (state = 0; state < vector_length(mfsm); ++state) {
		ffsm[state] = vector_create(sizeof(FINAL_ENTRY), 0, 16);
		
		for (jbp = 0; jbp < 16; ++jbp) {
			ffsm[state][jbp].next_state = mfsm[state][jbp].next_state;

			for (k = 0; k < 3; ++k) 
			  h[k] = mfsm[state][jbp].hamming[k];

			ffsm[state][jbp].hamming_increment = SET_H(h[0],h[1],h[2]);
		}
	}

	destroy_master_decoder(mfsm);

	return(ffsm);
}

void destroy_final_decoder(FINAL_ENTRY **ffsm)
{
	int i;

	for (i = 0; i < vector_length(ffsm); ++i) 
	  vector_destroy(ffsm[i]);

	vector_destroy(ffsm);
}

void print_final_decoder(FINAL_ENTRY **ffsm) 
{
	int i, jbp;
	
	printf("%d states in final decoder \n", vector_length(ffsm));
	
	for (i = 0; i < vector_length(ffsm); ++i) {
		printf("State %3d\n", i);
		for (jbp = 0; jbp < 16; ++jbp) {
			printf("\t%1X: %d %d %d  %4d\n", jbp,
			       SELECT_H1(ffsm[i][jbp].hamming_increment),
			       SELECT_H2(ffsm[i][jbp].hamming_increment),
			       SELECT_H3(ffsm[i][jbp].hamming_increment), 
			       ffsm[i][jbp].next_state);
		}
	}
}
