
/* DNA Compression Program -- Peter N. Yianilos 1996.
   Part of a joint project with David Loewenstern. */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <limits.h>
#include <ctype.h>

#include "util.h"
#include "amino.h"

#define MAXCONTEXT 256

void srand48(long seedval);
double drand48(void);

static FINAL_ENTRY **fsm;

void usage(char *name) 
{
	fprintf(stderr,
		"Usage is: %s <train> <#smp1> <seed> <test> <#smp2> <#em> <metric> ",
		name);
	
	fprintf(stderr, "<alphabetfile> <size1> [<size2>...]\n\n");

	fprintf(stderr, "\t<trainfile>,<testfile> ... ASCII files containing the DNA from\n");
	fprintf(stderr, "\t     which to train and test -- consisting of 'a','c','g','t'\n");
	fprintf(stderr, "\t     and whitespace.\n\n");

	fprintf(stderr, "\t<#smp1/2> ... Integer giving the number of random samples to\n");
	fprintf(stderr, "\t     select from the training/test file.  <#smp1> limits\n");
	fprintf(stderr, "\t     time and memory for training.  <#smp2> limits test time.\n\n");
	
	fprintf(stderr, "\t<seed> ... Integer providing random seed.\n\n");

	fprintf(stderr, "\t<#em> ... The number of EM iterations to perform\n\n");

	fprintf(stderr, "\t<metric> ... The syntax is a/n:f/m.  Specifying a:m corresponds\n");
	fprintf(stderr, "\t     to amino-hamming-distance with multiple directions, i.e. \n");
	fprintf(stderr, "\t     forward as well as reverse-complement matching attempts.\n");
	fprintf(stderr, "\t     Specifying n:f corresponds to nucleotide-hamming-distance\n");
	fprintf(stderr, "\t     with forward matching only.  All four possibilities may\n");
	fprintf(stderr, "\t     be used: a:f, a:m, n:f, n:m.\n\n");

	fprintf(stderr, "\t<size#> ... Each such term specifies context sizes to be used\n");
	fprintf(stderr, "\t     in the model.  Each term may given in one three forms: \n");
	fprintf(stderr, "\t     \"%%d\", \"%%d-%%d\", or, \"%%d-%%d:%%d\".  The first form\n");
	fprintf(stderr, "\t     specifies a single context length, the second a range, and\n");
	fprintf(stderr, "\t     the third a range with an increment.\n\n");
	
	fprintf(stderr, "\t<alphabetfile> ... First line gives alphabet size.  The\n");
	fprintf(stderr, "\t     characters beginning on the following line define the\n");
	fprintf(stderr, "\t     alphabet.  Any byte value is legal\n\n");
	
	fprintf(stderr, "\t     Note: There is built-in support for cross-validation.  If\n");
	fprintf(stderr, "\t     <testfile> is given as '@8:3' then the training file is\n");
	fprintf(stderr, "\t     divided into 8 roughly equal segments.  The third is\n");
	fprintf(stderr, "\t     deleted and becomes the test set.\n\n");
	
	exit(-1);
}

typedef struct model MODEL;

void initialize_alphabet(char *file);
char *read_dna(char *file);
int *random_permutation(int n);
MODEL *learn_model(char *train, int *context_sizes, int *train_permutation,
		   int n_em, int metric);
void test_model(MODEL *model, char *train, char *test, int *permutation,
		int *context_sizes, int metric);
void destroy_model(MODEL *model);

#define AMINO_DISTANCE      1
#define MULTIPLE_DIRECTIONS 2

int Alphabet_size, Alphabet[256], Ignore_space;

int main(int argn, char *argv[])
{
	int i, arg, from, to, inc, *context_sizes, size, max_context = 0;
	int xway, xnum, xinc, xfrom, xto;
	int ntrsamples, ntesamples;
	int *trpermutation, *tepermutation;
	int n_em, metric;
	char *train, *test, *temp, bp;
	MODEL *model;
	
	/* Begin processing command line arguments */

	if (argn < 8) 
	  usage(argv[0]);

	/* Initialize random seed so that even zero generates a non-trivial seed */ 

	srand48((long)(LONG_MAX * drand48()) ^ atol(argv[3]));
	
	/* Set number of samples */

	if ((ntrsamples = atoi(argv[2])) <= 0) {
		fprintf(stderr, "Bad #samples field: %s\n", argv[2]);
		usage(argv[0]);
	}

	if ((ntesamples = atoi(argv[5])) <= 0) {
		fprintf(stderr, "Bad #samples field: %s\n", argv[5]);
		usage(argv[0]);
	}
	
	/* Handle the #EM iterations specification */

	if ((n_em = atoi(argv[6])) < 0) {
		fprintf(stderr, "Bad #EM iterations: %s\n", argv[6]);
		usage(argv[0]);
	}
	
	/* Process the alphabet specification */

	initialize_alphabet(argv[8]);

	printf("Alphabet size = %d, Ignore_space = %d\n", Alphabet_size, Ignore_space);
	for (i = 0; i < 256; ++i)
	  if (Alphabet[i] >= 0)
	    printf("\t%02X: %d\n", i, Alphabet[i]);

	/* Handle the metric specification */

	if (strcmp(argv[7], "n:f") == 0) {
		metric = 0;
	} else if (strcmp(argv[7], "n:m") == 0) {
		assert(Alphabet_size == 4);
		metric = MULTIPLE_DIRECTIONS;
	} else if (strcmp(argv[7], "a:f") == 0) {
		assert(Alphabet_size == 4);
		metric = AMINO_DISTANCE;
	} else if (strcmp(argv[7], "a:m") == 0) {
		assert(Alphabet_size == 4);
		metric = AMINO_DISTANCE+MULTIPLE_DIRECTIONS;
	} else {
		fprintf(stderr, "Bad <metric> field: %s\n", argv[7]);
		usage(argv[0]);
	}

	/* Process the context length specification arguments */
	
	context_sizes = vector_create(sizeof(int), 0, MAXCONTEXT+1);

	for (size = 0; size <= MAXCONTEXT; ++size)
	  context_sizes[size] = 0;
	
	for (arg = 9; arg < argn; ++arg) {
		
		/* Parse all three allowed forms into common information */

		if (sscanf(argv[arg], "%d-%d:%d", &from, &to, &inc) == 3) {
			;
		} else if (sscanf(argv[arg], "%d-%d", &from, &to) == 2) {
			inc = 1;
		} else if (sscanf(argv[arg], "%d", &from) == 1) {
			to = from;
			inc = 1;
		} else {
			fprintf(stderr, "Bad size specification syntax: %s\n", argv[arg]);
			usage(argv[0]);
		}
		
		if (inc == 0) {
			fprintf(stderr, "Zero increment is illegal: %s\n", argv[arg]);
			usage(argv[0]);
		}

		for (size = from; size <= to; size += inc) {
			if (size < 1 || size > MAXCONTEXT || context_sizes[size]) {
				fprintf(stderr, "Error involving size specification: %s\n",
					argv[arg]);
				usage(argv[0]);
			}
			context_sizes[size] = 1;
		}
	}

	/* Read training and test DNA data */

	if (sscanf(argv[4], "@%d:%d", &xway, &xnum) == 2) {
		temp = read_dna(argv[1]);

		if ((xway <= 1) || (xnum < 1) || (xnum > xway) ||
		    (xway > vector_length(temp))) {
			fprintf(stderr, "Bad cross validation specification: %s\n", argv[4]);
			usage(argv[0]);
		}

		xinc = vector_length(temp) / xway;
		
		xfrom = (xnum-1) * xinc;
		
		if (xnum < xway) 
		  xto = xnum * xinc - 1;
		else
		  xto = vector_length(temp) - 1;
		
		train = vector_create(sizeof(char), 0, 0);
		test  = vector_create(sizeof(char), 0, 0);

		for (i = 0; i < vector_length(temp); ++i) {
			bp = temp[i];
			if (i >= xfrom && i <= xto)
			  test = vector_extend(test, &bp);
			else
			  train = vector_extend(train, &bp);
		}
		
		vector_destroy(temp);
	} else {
		train = read_dna(argv[1]);
		test  = read_dna(argv[4]);
	}

	printf("Program Arguments:\n\t");
	for (i = 0; i < argn; ++i) 
	  printf("%s ", argv[i]);
	printf("\n");

	printf("Training set '%s' contains %d base pairs\n",
	       argv[1], vector_length(train));
	printf("Testing  set '%s' contains %d base pairs\n",
	       argv[4], vector_length(test));

	if (ntrsamples > vector_length(train)) {
		fprintf(stderr, "More samples requested than available training data: %s\n", 
			argv[2]);
		usage(argv[0]);
	}
	  
	if (ntesamples > vector_length(test)) {
		fprintf(stderr, "More samples requested than available testing data: %s\n", 
			argv[5]);
		usage(argv[0]);
	}

	for (size = 1; size <= MAXCONTEXT; ++size)
	  if (context_sizes[size])
	    max_context = size;
	
	if (vector_length(train) < max_context+1)
	  error_fatal("main", "Train file is too short");

	printf("Random Seed = %s\n", argv[3]);

	printf("Hamming distance: %s : %s\n", metric&AMINO_DISTANCE? "amino": "nulcleotide",
	       metric&MULTIPLE_DIRECTIONS? "forward+reverse-complement": "forward");

	printf("%d Training samples will be used.\n", ntrsamples);
	printf("%d Testing samples will be used.\n",  ntesamples);
	printf("%d EM Iterations will be performed\n\n", n_em);

	/* Done processing arguments */

	/* Generate random sampling vectors */

	trpermutation = random_permutation(vector_length(train));
	trpermutation = vector_contract(trpermutation, vector_length(train) - ntrsamples);
	assert(vector_length(trpermutation) == ntrsamples);
	
	tepermutation = random_permutation(vector_length(test));
	tepermutation = vector_contract(tepermutation, vector_length(test) - ntesamples);
	assert(vector_length(tepermutation) == ntesamples);

	/* Train and then test */
	    
	model = learn_model(train, context_sizes, trpermutation, n_em, metric);
	test_model(model, train, test, tepermutation, context_sizes, metric); 
	    
	/* Clean up */
	
	destroy_model(model);
	vector_destroy(context_sizes);
	vector_destroy(train);
	vector_destroy(test);
	vector_destroy(trpermutation);
	vector_destroy(tepermutation);
	
	if (fsm)
	  destroy_final_decoder(fsm);

	/* Check for memory leaks */

	assert(mem_total_space() == 0);

	return(0);
}

void initialize_alphabet(char *file)
{
	FILE *in;
	int i, c;
	
	if ((in = fopen(file, "r")) == NULL) 
	  error_fatal("initialize_alphabet", "Cannot open file %s", file);
	
	Alphabet_size = -1;
	fscanf(in, "%d\n", &Alphabet_size);
	if (Alphabet_size < 2 || Alphabet_size > 256)
	  error_fatal("initialize_alphabet", "Alphabet size %d out of range", Alphabet_size);
	  
	for (i = 0; i < 256; ++i)
	  Alphabet[i] = -1;

	for (i = 0; i < Alphabet_size; ++i) {
		if ((c = fgetc(in)) == EOF)
		  error_fatal("initialize_alphabet", "Unexpected EOF");
		if (Alphabet[c] == -1)
		  Alphabet[c] = i;
		else
		  error_fatal("initialize_alphabet",
			      "Alphabet member %02X already defined", c);
	}

	while ((c = fgetc(in)) != EOF)
	  if (!isspace(c))
	    error_fatal("initialize_alphabet", "Non-whitespace after all chars defined!");

	fclose(in);

	if ((Alphabet[' '] == -1) && (Alphabet['\t'] == -1) && (Alphabet['\n'] == -1))
	  Ignore_space = 1;
}

char *read_dna(char *file) 
{
	char bp, *v;
	int c;
	FILE *ioptr;

	v = vector_create(sizeof(char), 0, 0);

	if ((ioptr = fopen(file, "r")) == NULL)
	  error_fatal("main", "Cannot open file: %s", file);
	
	while ((c = getc(ioptr)) != EOF) {

		if (Ignore_space && isspace(c))
		  continue;

		if (Alphabet[c] >= 0)
		  bp = Alphabet[c];
		else
		  error_fatal("read_dna", "Illegal character %c in file %s", c, file);
		
		v = vector_extend(v, &bp);
	}

	fclose(ioptr);
	
	return(v);
}

double bits(double p)
{
	return(-log(p) / log(2.0));
}

/* Routines to extract a random sample without replacement */

typedef struct {
	double value;
	int index;
} SLIST;

static int slist_compare(SLIST *a, SLIST *b)
{
	if (a->value < b->value) 
	  return(1);
	else if (b->value < a->value) 
	  return(-1);
	else 
	  return(a->index - b->index);
}

int *random_permutation(int n)
{
	int i, *permutation;
	SLIST *list;

	list = vector_create(sizeof(SLIST), 0, n);

	for (i = 0; i < n; ++i) {
		list[i].value = drand48();
		list[i].index = i;
	}

	qsort(list, n, sizeof(SLIST), (int (*)())slist_compare);
 
	permutation = vector_create(sizeof(int), 0, n);
 	
	for (i = 0; i < n; ++i)
	  permutation[i] = list[i].index;

	vector_destroy(list);

	return(permutation);
}

/* 
  The following two routines normalize vectors of floats and doubles
  respectively.  Before doing so the float version adds a small
  constant FLATTEN to each entry.  This is in a sense good statistical
  hygene, but beyond this, it solves two very real problems in our
  application:
  
  1. It is possible (but not very likely) that for some window size "w", the count
  corresponding to an alphabet member (ACGT for DNA) may be zero for *every* hamming
  distance.  In this case the model corresponding to this window size would 
  predict this symbol with probability zero, i.e. infinite bits.  Beyond this it is
  possible that this is true for every window size.  In this case the overall model
  would suffer from the same problem.  Adding FLATTEN ensures that every symbol
  is coded in a finite number of bits.
  
  2. It is possible (but not very likely) that for some sample, we'll see a hamming
  distance "h" match in the past, but for some g>h have no matches.  Here the
  FLATTEN addition will result in a flat (max entropy) predictor for distance g.
  Note that normalize_double does not flatten, and must not.  See create_model().

  The double version includes a special case in which a zero vector is normalized
  to itself. This is required because we may encounter zero expectation
  accumlator vectors in training.  
  */

#define FLATTEN 1E-10

void normalize_float(float *x)
{
	int i;
	double t;

	for (i = 0, t = 0; i < vector_length(x); ++i) {
		x[i] += FLATTEN;
		t += x[i];
	}
	
	t = 1/t;

	for (i = 0; i < vector_length(x); ++i) 
	  x[i] *= t;
}

double normalize_double(double *x)
{
	int i, n;
	double t;

	n = vector_length(x);

	for (i = 0, t = 0; i < n; ++i) 
	  t += x[i];
	
	if (t == 0) 
	  return(0);

	t = 1/t;

	for (i = 0; i < n; ++i) 
	  x[i] *= t;

	return(1/t);
}

typedef struct sample SAMPLE;

void extract_context(char *input, int position, char *output, int length);
SAMPLE *gather_stats(char *input, int *context_sizes, char *context, int avoid, int metric);
void destroy_stats(SAMPLE *s);
void stat_summary(SAMPLE **list, int length);
void set_first_hamming(SAMPLE *s);
void make_probabilities(SAMPLE *s);
MODEL *create_model(SAMPLE *s);
void print_model(MODEL *m);
void sanity_check_model(MODEL *m, int *context_sizes);
double evaluate_estep(SAMPLE *s, MODEL *m, int do_estep);
void sanity_check_expectations(MODEL *m, int n_train);
void maximize_model(MODEL *m);

MODEL *learn_model(char *train, int *context_sizes, int *train_permutation,
		   int n_em, int metric)
{
	int i, em, n_train;
	int from, to, inc;
	SAMPLE **samples;
	char context[MAXCONTEXT];
	MODEL *model;
	double total_bits;
	
	/* Build statistics structure for each random training sample */

	n_train = vector_length(train_permutation);
	samples = vector_create(sizeof(SAMPLE *), 0, n_train);
	
	for (i = 0; i < n_train; ++i) {
		extract_context(train, train_permutation[i], context, MAXCONTEXT);
		samples[i] = gather_stats(train, context_sizes, context,
					  train_permutation[i], metric);
	}
	
	/* Report on summary statistics */

	if (metric == 0)
	  stat_summary(samples, vector_length(train));
	else
	  printf("Statistics summary generated only for n:f metric mode\n");

	/* Fill in first_hamming vector in each sample structure */

	for (i = 0; i < n_train; ++i) 
	  set_first_hamming(samples[i]);

	/* Now convert the counts to a flattened probability function */

	for (i = 0; i < n_train; ++i) 
	  make_probabilities(samples[i]);

	/* Create the initial model and perform sanity check */

	model = create_model(samples[0]);
	sanity_check_model(model, context_sizes);

	/* Use EM to improve the initial model */

	for (em = 1; em <= n_em; ++em) {
		printf("Starting EM iteration %d\n", em);

		/* Alternate scan direction to improve virual memory LRU performance */

		if (em%2) {
			from = n_train-1;
			to = -1;
			inc = -1;
		} else {
			from = 0;
			to = n_train;
			inc = 1;
		}

		for (i = from, total_bits = 0; i != to; i += inc)
		  total_bits += bits(evaluate_estep(samples[i], model, 1));

		sanity_check_expectations(model, n_train);
		printf("Expectation step complete, avg bits = %10g\n", total_bits / n_train);
		maximize_model(model);
		printf("Maximization complete\n");
	}
	
	for (i = 0, total_bits = 0; i < n_train; ++i)
	  total_bits += bits(evaluate_estep(samples[i], model, 0));
	printf("Final avg bits .................... = %10g\n", total_bits / n_train);

	/* Print the model to the log file */

	print_model(model);

	/* Cleanup */
	
	printf("Approximate Peak Memory Use: %.2fMB\n", mem_total_space()/1E6);

	for (i = 0; i < vector_length(train_permutation); ++i) 
	  destroy_stats(samples[i]);
	  
	vector_destroy(samples);

	return(model);
}

/* Extract context from input starting at position and working backwards (towards
   the beginning of the input string.  Output[0] is input[position], Output[1]
   is input[position-1], and so on.  When the context extends beyond the start of
   the input string, the output is assigned "-1". */

void extract_context(char *input, int position, char *output, int length)
{
	int i;

	for (i = 0; i < length; ++i) {
		if (position-i >= 0)
		  output[i] = input[position-i];
		else
		  output[i] = -1;
	}
}

struct sample {
	char predicted_base_pair;
	float ***stats;             /* stats[window_length][hamming_distance][0-3] */
	int *first_hamming;
};

void make_probabilities(SAMPLE *s)
{
	int w, hd, maxwindow;

	maxwindow = vector_length(s->stats) - 1;
	for (w = 1; w <= maxwindow; ++w) {
		if (s->stats[w] == NULL) 
		  continue;
		for (hd = 0; hd <= w; ++hd) 
		  normalize_float(s->stats[w][hd]);
	}
}

void set_first_hamming(SAMPLE *s)
{
	int w, hd, maxwindow, t, a;

	maxwindow = vector_length(s->stats) - 1;
	for (w = 1; w <= maxwindow; ++w) {
		if (s->stats[w] == NULL) {
			s->first_hamming[w] = -1;
			continue;
		}
		for (hd = 0, t = 0; hd <= w; ++hd) {
			for (a = 0; a < vector_length(s->stats[w][hd]); ++a) 
			  t += s->stats[w][hd][a];
			if (t)
			  break;
		}
		assert(hd <= w);
		s->first_hamming[w] = hd;
	}
}

void scan_text_n(float ***stats, char *input, int *context_sizes, char *context,
		 int avoid, int metric);

void scan_text_a(float ***stats, char *input, int *context_sizes, char *context,
		 int avoid, int metric);

SAMPLE *gather_stats(char *input, int *context_sizes, char *context, int avoid, int metric)
{
	SAMPLE *s;
	int i, j, k, max_window = 0;
	
	/* Allocate and initialize the structure */

	s = mem_alloc(sizeof(SAMPLE));

	s->predicted_base_pair = context[0];

	for (i = 0; i < vector_length(context_sizes); ++i) 
	  if (context_sizes[i])
	    max_window = i;

	s->stats = vector_create(sizeof(float **), 0, max_window+1);
	s->first_hamming = vector_create(sizeof(int), 0, max_window+1);

	for (i = 1; i <= max_window; ++i) {
		if (context_sizes[i] == 0) {
			s->stats[i] = NULL;
			continue;
		}
		s->stats[i] = vector_create(sizeof(float *), 0, i+1);
		for (j = 0; j < i+1; ++j) {
			s->stats[i][j] = vector_create(sizeof(float), 0, Alphabet_size);
			for (k = 0; k < Alphabet_size; ++k)
			  s->stats[i][j][k] = 0;
		}
	}

	/* Fill in the statistics */
	
	if (metric & AMINO_DISTANCE)
	  scan_text_a(s->stats, input, context_sizes, context, avoid, metric);
	else 
	  scan_text_n(s->stats, input, context_sizes, context, avoid, metric);

	return(s);
}

void complement(char *context, int length);

void scan_text_n(float ***stats, char *input, int *context_sizes, char *context,
		  int avoid, int metric)
{
	int head, max_window, w, next_bp, val, hd, ninput;
	
	max_window = vector_length(stats) - 1;
	ninput = vector_length(input);

	for (head = ninput - 1; head >= 0; --head) {
		if (head == avoid)
		  continue;
		next_bp = input[head];
		hd = 0;
		for (w = 1; w <= max_window; ++w) {
			if (head-w >= 0)
			  val = input[head-w];
			else
			  break;
			if (val != context[w])
			  ++hd;
			if (context_sizes[w])
			  ++stats[w][hd][next_bp];
		}
	}

	if (metric == 0)
	  return;

	/* Look for reversed and complemented matches */

	complement(context, MAXCONTEXT);
	
	for (head = ninput - 1; head >= 0; --head) {
		if (head == avoid)
		  continue;
		next_bp = 3-input[head];
		assert(next_bp >= 0 && next_bp <= 3);
		hd = 0;
		for (w = 1; w <= max_window; ++w) {
			if (head+w < ninput)
			  val = input[head+w];
			else
			  break;
			if (val != context[w])
			  ++hd;
			if (context_sizes[w])
			  ++stats[w][hd][next_bp];
		}
	}
}

#define MIN3(x,y,z) ((x)<(y)?((x)<(z)?(x):(z)):((y)<(z)?(y):(z)))

void scan_text_a(float ***stats, char *input, int *context_sizes, char *context,
		  int avoid, int metric)
{
	int head, max_window, w, next_bp, val, hd, ninput;
	int state, vhd[3], hacc, jbp;
	
	if (fsm == NULL) 
	  fsm = build_final_decoder();

	max_window = vector_length(stats) - 1;
	ninput = vector_length(input);

	for (head = ninput - 1; head >= 0; --head) {
		if (head == avoid)
		  continue;
		next_bp = input[head];
		hacc = 0;
		state = 0;
		for (w = 1; w <= max_window; ++w) {
			if (head-w >= 0)
			  jbp = input[head-w]*4 + context[w];
			else
			  break;

			hacc += fsm[state][jbp].hamming_increment;
			state = fsm[state][jbp].next_state;

			if (context_sizes[w]) {
				vhd[0] = SELECT_H1(hacc);
				vhd[1] = SELECT_H2(hacc);
				vhd[2] = SELECT_H3(hacc);

				hd = MIN3(vhd[0],vhd[1],vhd[2]);

				++stats[w][hd][next_bp];
			}
		}
	}

	if (metric == AMINO_DISTANCE)
	  return;

	/* Look for reversed and complemented matches */

	complement(context, MAXCONTEXT);
	
	for (head = ninput - 1; head >= 0; --head) {
		if (head == avoid)
		  continue;
		next_bp = 3-input[head];
		hacc = 0;
		state = 0;
		for (w = 1; w <= max_window; ++w) {
			if (head+w < ninput)
			  jbp = input[head+w]*4 + context[w];
			else
			  break;
			
			hacc += fsm[state][jbp].hamming_increment;
			state = fsm[state][jbp].next_state;

			if (context_sizes[w]) {
				vhd[0] = SELECT_H1(hacc);
				vhd[1] = SELECT_H2(hacc);
				vhd[2] = SELECT_H3(hacc);

				hd = MIN3(vhd[0],vhd[1],vhd[2]);
				
				++stats[w][hd][next_bp];
			}
		}
	}
}

double factorial(int n)
{
	double f = 1;

	assert(n >= 0);

	for ( ; n; --n)
	  f *= n;

	return(f);
}

void stat_summary(SAMPLE **list, int length)
{
	int w, hd, item, i, n_hits, pbp;
	double expect, actual, check, t, exp_correct;
	
	for (w = 1; w <= vector_length(list[0]->stats)-1; ++w) {
		
		printf("Window size %2d: ", w);

		if (list[0]->stats[w] == NULL) {
			printf("Statistics not gathered\n");
			continue;
		}

		printf("\n");

		check = 2 * (length - (w+1));

		for (hd = 0; hd <= w; ++hd) {
			printf("\tHamming %2d ", hd);
			
			expect = (length - (w+1)) *
			  pow(1.0/Alphabet_size,(double)w-hd) *
			  pow((Alphabet_size-1.0)/Alphabet_size,(double)hd) * 
			  factorial(w) / (factorial(hd) * factorial(w-hd));

			printf("Expect %11g ", expect);
			check -= expect;
			
			n_hits = 0;
			exp_correct = 0;
			
			for (item = 0, actual = 0; item < vector_length(list); ++item) {
				for (i = 0, t = 0; i < Alphabet_size; ++i) {
					t += list[item]->stats[w][hd][i];
				}
				if (t != 0) {
					++n_hits;
					pbp = list[item]->predicted_base_pair;
					exp_correct += list[item]->stats[w][hd][pbp] / t;
				}
				actual += t;
			}
			
			actual /= vector_length(list);

			printf("Actual %10g ", actual);
			check -= actual;
			
			printf("Hit-Rate %7g%% ", (100.0 * n_hits) / vector_length(list));

			printf("E[Pr(true_bp|hit)] ");
			if (n_hits)
			  printf("%7g%% ", 100 * exp_correct / n_hits);
			else
			  printf("n/a ");
			  
			printf("\n");
		}
		assert(check < 1E-3 && check > -(2+1E-3));
		/* Note that check will be zero, except when some training samples lie
		   very near to the beginning of the training data. In this clase it
		   can swing slightly negative */
	}
}

void destroy_stats(SAMPLE *s)
{
	int i, j;

	for (i = 1; i < vector_length(s->stats); ++i) {
		if (s->stats[i] == NULL)
		  continue;
		for (j = 0; j < vector_length(s->stats[i]); ++j) {
			if (s->stats[i][j])
			  vector_destroy(s->stats[i][j]);
		}
		vector_destroy(s->stats[i]);
	}

	vector_destroy(s->stats);
	vector_destroy(s->first_hamming);
	mem_free(s);
}

struct model {
	double *choose_window; /* choose_window[w] */
	double *expect_window;
	double ***choose_hd;   /* choose_hd[w][first_hamming][first_hamming <= hd <= w] */
	double ***expect_hd;
};

/* Create and initialize the model.  Note that the window choice parameters should
not be flattened when maximized because a parameter value of zero is used to indicate
that a particular window size is not present in the model */

MODEL *create_model(SAMPLE *s)
{
	int w, n_windows, maxwin, fh, hd;
	MODEL *m;
	
	/* Allocate the model structure */

	m = mem_alloc(sizeof(MODEL));
	
	/* Allocate and initialize the window selection parameters */

	maxwin = vector_length(s->stats) - 1;

	m->choose_window = vector_create(sizeof(double), 0, maxwin+1);
	m->expect_window = vector_create(sizeof(double), 0, maxwin+1);

	for (w = 1, n_windows = 0; w <= maxwin; ++w) 
	  if (s->stats[w] != NULL) 
	    ++n_windows;
		
	for (w = 1; w <= maxwin; ++w) {
		if (s->stats[w] == NULL) 
		  m->choose_window[w] = 0;
		else 
		  m->choose_window[w] = 1.0 / n_windows;
		m->expect_window[w] = 0;
	}

	m->expect_window[0] = 0;

	/* Allocate and initialize the submodel parameters */

	m->choose_hd = vector_create(sizeof(double **), 0, maxwin+1);
	m->expect_hd = vector_create(sizeof(double **), 0, maxwin+1);

	for (w = 1; w <= maxwin; ++w) {
		m->choose_hd[w] = vector_create(sizeof(double *), 0, w+1);
		m->expect_hd[w] = vector_create(sizeof(double *), 0, w+1);
		for (fh = 0; fh <= w; ++fh) {
			m->choose_hd[w][fh] = vector_create(sizeof(double), 0, w+1);
			m->expect_hd[w][fh] = vector_create(sizeof(double), 0, w+1);
			for (hd  = fh; hd <= w; ++hd) 
			  m->choose_hd[w][fh][hd] = 1.0 / (w + 1 - fh);
			for (hd  = 0; hd <= w; ++hd) 
			  m->expect_hd[w][fh][hd] = 0;
		}
	}

	return(m);
}

void print_model(MODEL *m)
{
	int w, fh, hd;

	printf("\nMODEL Parameters:\n");

	for (w = 1; w < vector_length(m->choose_window); ++w) {
		if (m->choose_window[w] == 0)
		  continue;
		printf("\tWindow size %2d: %10g\n", w, m->choose_window[w]);
		for (fh = 0; fh <= w; ++fh) {
			printf("\t\tFirst Hamming %2d\n", fh);
			for (hd = fh; hd <= w; ++hd) {
				printf("\t\t\tDistance %2d: %10g\n",
				       hd, m->choose_hd[w][fh][hd]);
			}
		}
	}
}

void destroy_model(MODEL *m)
{
	int w, fh;

	vector_destroy(m->choose_window);
	vector_destroy(m->expect_window);

	for (w = 1; w < vector_length(m->choose_hd); ++w) {
		for (fh = 0; fh <= w; ++fh) {
			vector_destroy(m->choose_hd[w][fh]);
			vector_destroy(m->expect_hd[w][fh]);
		}
		vector_destroy(m->choose_hd[w]);
		vector_destroy(m->expect_hd[w]);
	}

	vector_destroy(m->choose_hd);
	vector_destroy(m->expect_hd);
	
	mem_free(m);
}

double evaluate_estep(SAMPLE *s, MODEL *m, int do_estep)
{
	int w, maxwin, hd, first_hamming, pbp;
	double *mid, prob, *submod;
	float **wc;
	
	maxwin = vector_length(s->stats) - 1;

	/* Build vector of probabilities conditioned on each window size */

	mid = vector_create(sizeof(double), 0, maxwin+1);

	for (w = 1; w <= maxwin; ++w) {
		if (s->stats[w] == NULL)
		  continue;

		first_hamming = s->first_hamming[w];
		submod = m->choose_hd[w][first_hamming];
		pbp = s->predicted_base_pair;
		wc = s->stats[w];
		
		for (hd = first_hamming, prob = 0; hd <= w; ++hd) 
		  prob += submod[hd] * wc[hd][pbp];

		mid[w] = prob;
	}
	
	/* Compute overall sample probability by mixing window sizes together */

	for (w = 1, prob = 0; w <= maxwin; ++w) {
		if (s->stats[w] == NULL)
		  continue;

		prob += m->choose_window[w] * mid[w];
	}

	/* Perform expectation step if requested */

	if (do_estep) {

		/* Accumulate window selection expecations */

		for (w = 1; w <= maxwin; ++w) {
			if (s->stats[w] == NULL)
			  continue;
			
			m->expect_window[w] += m->choose_window[w] * mid[w] / prob;
		}

		/* Accumulate submodel expectations */
		
		for (w = 1; w <= maxwin; ++w) {
			if (s->stats[w] == NULL)
			  continue;

			first_hamming = s->first_hamming[w];
			submod = m->choose_hd[w][first_hamming];
			pbp = s->predicted_base_pair;
			wc = s->stats[w];
			
			for (hd = first_hamming; hd <= w; ++hd) {
				m->expect_hd[w][first_hamming][hd] += 
				  m->choose_window[w] * submod[hd] * wc[hd][pbp] / prob;
			}
		}
	}
	
	/* Clean up and return */

	vector_destroy(mid);

	return(prob);
}

void sanity_check_expectations(MODEL *m, int n_train)
{
	int w, maxwin, first_hamming, hd;
	double total_window, total_submod;
	
	maxwin = vector_length(m->choose_window)-1;

	/* Check window selection expectations */

	for (w = 1, total_window = 0; w <= maxwin; ++w) 
	  total_window += m->expect_window[w];

	assert(fabs(total_window - n_train) < 1E-2);

	/* Check submodel weighting expecations */
	
	for (w = 1, total_submod = 0; w <= maxwin; ++w) {
		if (m->choose_window[w] == 0)
		  continue;
		for (first_hamming = 0; first_hamming <= w; ++first_hamming) 
		  for (hd = first_hamming; hd <= w; ++hd) 
		    total_submod += m->expect_hd[w][first_hamming][hd];
	}

	assert(fabs(total_submod - n_train) < 1E-2);
}

void maximize_model(MODEL *m)
{
	int w, maxwin, first_hamming, hd;

	maxwin = vector_length(m->choose_window) - 1;

	/* Maximize window selection parameters */

	normalize_double(m->expect_window);

	for (w = 1; w <= maxwin; ++w) {
		m->choose_window[w] = m->expect_window[w];
		m->expect_window[w] = 0;
	}
	
	/* Maximize submodel parameters */

	for (w = 1; w <= maxwin; ++w) {
		if (m->choose_window[w] == 0)
		  continue;
		for (first_hamming = 0; first_hamming <= w; ++first_hamming) {
			if (normalize_double(m->expect_hd[w][first_hamming]) == 0) {
				for (hd = first_hamming; hd <= w; ++hd)
				  m->expect_hd[w][first_hamming][hd] =
				    1.0 / (w + 1 - first_hamming);
			}
			for (hd = first_hamming; hd <= w; ++hd) {
				m->choose_hd[w][first_hamming][hd] =
				  m->expect_hd[w][first_hamming][hd];
				m->expect_hd[w][first_hamming][hd] = 0;
			}
		}
	}
}

void sanity_check_model(MODEL *m, int *context_sizes)
{
	int w, fh;
	SAMPLE *s;
	char *null;
	double prob;

	null = vector_create(sizeof(char), 0, 0);

	for (fh = 0; fh < vector_length(m->choose_window); ++fh) {
		s = gather_stats(null, context_sizes, "\000", 0, 0);

		for (w = 1; w < vector_length(s->first_hamming); ++w) 
		  s->first_hamming[w] = (fh <= w)? fh: w;
	  
		make_probabilities(s);
		
		prob = evaluate_estep(s, m, 0);
		assert(fabs(1/prob - Alphabet_size) < 1E-3);

		destroy_stats(s);
	}

	vector_destroy(null);
}

int verify_probability_function(SAMPLE *s, MODEL *m) 
{
	int bp;
	double prob;
	
	for (bp = 0, prob = 0; bp < Alphabet_size; ++bp) {
		s->predicted_base_pair = bp;
		prob += evaluate_estep(s, m, 0);
	}

	if (fabs(prob - 1) > 1E-3)
	  return(0);
	else
	  return(1);
}

/* Note that one may specify the same file as train and test if a random sample is
used for both cases.  The test_model() routine includes special code for this case */

void test_model(MODEL *model, char *train, char *test, int *permutation,
		int *context_sizes, int metric)
{
	int i, same, avoid;
	SAMPLE *s;
	char context[MAXCONTEXT];
	double sample_bits, total_bits;

	/* Set flag if train == test */

	if (vector_length(train) == vector_length(test)) {
		for (i = 0; i < vector_length(train); ++i)
		  if (train[i] != test[i]) 
		    break;
		if (i < vector_length(test))
		  same = 0;
		else
		  same = 1;
	} else {
		same = 0;
	}
	
	/* Perform test */

	for (i = 0, total_bits = 0; i < vector_length(permutation); ++i) {
		avoid = same? permutation[i]: -1;
		extract_context(test, permutation[i], context, MAXCONTEXT);
		s = gather_stats(train, context_sizes, context, avoid, metric);
		
		set_first_hamming(s);
		make_probabilities(s);

		sample_bits = bits(evaluate_estep(s, model, 0));
		total_bits += sample_bits;
		
		printf("%8d: %8d %10g\n", i, permutation[i], sample_bits);
		
		assert(verify_probability_function(s, model));

		destroy_stats(s);
	}

	printf("Average bits = %10g\n", total_bits / vector_length(permutation));
}

/* Routine to complement a nucleotide context sequence */

void complement(char *context, int length)
{
	int i, t;
	
	for (i = 1; i < length; ++i) {	
		switch (context[i]) {
		      case 0:
			context[i] = 3;
			break;
		      case 1:
			context[i] = 2;
			break;
		      case 2:
			context[i] = 1;
			break;
		      case 3:
			context[i] = 0;
			break;
		      case -1:
			context[i] = -1;
			break;
		      default:
			error_fatal("reverse_complement",
				    "Bad character %d in DNA\n", context[i]);	
		}
	}
}
