
/* Program to test amino distance automaton -- Peter N. Yianilos 1996.
   Part of a joint project with David Loewenstern. */

#include <stdio.h>
#include <assert.h>

#include "util.h"
#include "amino.h"

char table[64][5] = {
	"aaaK", "aacN", "aagK", "aatN", "acaT", "accT", "acgT", "actT",
	"agaR", "agcS", "aggR", "agtS", "ataI", "atcI", "atgM", "attI",
	"caaQ", "cacH", "cagQ", "catH", "ccaP", "cccP", "ccgP", "cctP",
	"cgaR", "cgcR", "cggR", "cgtR", "ctaL", "ctcL", "ctgL", "cttL",
	"gaaE", "gacD", "gagE", "gatD", "gcaA", "gccA", "gcgA", "gctA",
	"ggaG", "ggcG", "gggG", "ggtG", "gtaV", "gtcV", "gtgV", "gttV",
	"taa-", "tacY", "tag-", "tatY", "tcaS", "tccS", "tcgS", "tctS", 
	"tga-", "tgcC", "tggW", "tgtC", "ttaL", "ttcF", "ttgL", "tttF",
};

int compare_codon(char *a, char *b)
{
	int i, j, k;

	for (i = 0; i < 64; ++i) {
		for (k = 0; k < 3; ++k) {
			if (a[k] == '?')
			  continue;
			if (a[k] != table[i][k])
			  break;
		}
		if (k == 3) {
			for (j = 0; j < 64; ++j) {
				for (k = 0; k < 3; ++k) {
					if (b[k] == '?')
					  continue;
					if (b[k] != table[j][k])
					  break;
				}
				if (k == 3) {
					if (table[i][3] == table[j][3])
					  return(0);
				}
			}
		}
	}

	return(1);
}

int compare_string(char *a, char *b, int ncodons)
{
	int i, h;

	for (i = 0, h = 0; i < ncodons; ++i) 
	  h += compare_codon(a+3*i, b+3*i);

	return(h);
}

void randbp(char *cbp, int *ibp)
{
	int bp;
	double drand48();

	bp = 4 * drand48();
	
	assert(bp >= 0 && bp <= 3);
	
	*ibp = bp;

	switch(bp) {
	      case 0:
		*cbp = 'a';
		break;
	      case 1:
		*cbp = 'c';
		break;
	      case 2:
		*cbp = 'g';
		break;
	      case 3:
		*cbp = 't';
		break;
	}
}

void randcodon(char *output)
{
	int i;

	randbp(output+0, &i);
	randbp(output+1, &i);
	randbp(output+2, &i);
	output[3] = 0;
}

int scancodon(char *output)
{
	int i, r;

	for (i = 0; i < 3; ++i) {
		while (1) {
			if ((r = fgetc(stdin)) == EOF)
			  return(0);
			if (!isspace(r))
			  break;
		}
		assert((r == 'a') || (r == 'c') || (r == 'g') || (r == 't'));
		output[i] = r;
	}
	output[3] = 0;
}

#define MAXLENGTH 48
#define NTEST 500

void main(int argn, char *argv[])
{
	int i, n, length, ncodons, h[3], fh[3], hacc, jbp, state;
	FINAL_ENTRY **fsm;
	int ai[MAXLENGTH], bi[MAXLENGTH];
	char a[MAXLENGTH+6], b[MAXLENGTH+6];
	
	/* If first argument is "filter" the program acts as a randomizer, reading
	   DNA from stdin, randomizing the assigned nucleotides so that the amino
	   acid sequence is preserved, and writing the result to stdout */

	if ((argn == 2) && (strcmp(argv[1], "filter") == 0)) {
		i = 0;
		while (scancodon(a)) {
			if (i++ % 1 == 0) {
				for (randcodon(b); compare_codon(a,b) == 1; randcodon(b))
				  ;
				printf("%s", b);
			} else {
				printf("%s", a);
			}
		}
		exit(0);
	}
	
	/* Initialization of automaton and "?" padded comparison strings */

	fsm = build_final_decoder();

	print_final_decoder(fsm);

	for (i = 0; i < MAXLENGTH+5; ++i) {
		a[i] = '?';
		b[i] = '?';
	}

	a[i] = 0;
	b[i] = 0;

	/* Main testing loop */
	
	for (length = 1; length <= MAXLENGTH; ++length) {
		
		printf("Length %2d\n", length);

		ncodons = 1 + (length+1)/3;

		for (n = 0; n < NTEST; ++n) {
			
			/* Draw random strings */
			
			for (i = 0; i < length; ++i) {
				randbp(a+2+i, ai+i);
				randbp(b+2+i, bi+i);
			}

			/* Compare them with the routines of this program */

			h[0] = compare_string(a+2, b+2, ncodons);
			h[1] = compare_string(a+1, b+1, ncodons);
			h[2] = compare_string(a+0, b+0, ncodons);
			
			/* Compare them with the finite state machine we're testing */

			for (i = 0, state = 0, hacc = 0; i < length; ++i) {
				jbp = ai[i]*4 + bi[i];
				assert(jbp >= 0 && jbp <= 15);
				hacc += fsm[state][jbp].hamming_increment;
				state = fsm[state][jbp].next_state;
			}

			fh[0] = SELECT_H1(hacc);
			fh[1] = SELECT_H2(hacc);
			fh[2] = SELECT_H3(hacc);

			for (i = 0; i < 3; ++i) 
			  if (h[i] != fh[i])
			    break;

			if (i != 3) {
				printf("Mismatch!\n");
				printf("a=%s\nb=%s\n", a, b);
				printf("Check Prog gives: %d %d %d\n", h[0], h[1], h[2]);
				printf("FSM gives:        %d %d %d\n", fh[0], fh[1], fh[2]);
			}
		}
	}

	/* Cleanup */

	destroy_final_decoder(fsm);
	assert(mem_total_space() == 0);
}
