
/* Extracted from common, mem, and vector libraries originally written
   by Kanzelberger, Ristad, and Yianilos. */

#include <assert.h>
#include <stdio.h>
#include <stdarg.h>

#include "util.h" 

void	*malloc(), *realloc(), *memcpy();

int Memory_bytes_used;

#define mem_header(mem_p)     ((memory_t *)(mem_p)-1)

/* Routine to allocate memory */
void *mem_alloc( int bytes )
{
  int	memsize;
  void	*p;

  assert(bytes >= 0);
  /* Allocate "bytes" plus header */
  memsize = sizeof(memory_t) + bytes;
  if ((p = malloc(memsize)) == NULL) {
    return NULL;
  }

  /* Store size in header, and keep track of
   * total memory allocated.
   */
  ((memory_t *)p)->self.bytes  = memsize;
  Memory_bytes_used 	      += memsize;

  p = (memory_t *)p + 1;	/* pointer to usable portion */
				
  return(p);
}


/* Routine to resize memory */
void *mem_realloc( void *pointer, int bytes )
{
  int		oldsize, newsize;
  memory_t	*header;

  assert(bytes >= 0);
  assert(pointer != NULL);

  header = (memory_t *)pointer - 1;

  oldsize = header->self.bytes;
  newsize = sizeof(memory_t) + bytes;

  if ((header = realloc(header, newsize)) == NULL) {
    return NULL;
  }
	
  header->self.bytes = newsize;
  Memory_bytes_used += (newsize - oldsize);

  return(header + 1);		/* pointer to usable portion */
}

/* Routine to free memory */
void mem_free( void *p )
{
  memory_t *h;

  if (p != NULL) {
    h = (memory_t *)p - 1;
    Memory_bytes_used -= h->self.bytes;
    free(h);
  }
}

/* Routine to move a block of memory from s2 to s1.  
 * Overlapping moves are handled correctly.
 */
void *mem_move( void *s1, void *s2, unsigned long length )
{
  unsigned char	*p1, *p2;

  assert(s1 != NULL);
  assert(s2 != NULL);

  p1 = s1;
  p2 = s2;

  if (s1 < s2) {
    while (length--) {
      *p1++ = *p2++;
    }

  } else if (s1 > s2) {
    p1 += length;
    p2 += length;

    while (length--) {
      *--p1 = *--p2;
    }
  }

  return s1;
}

#define ALLOCN(n)		 ((n)*1)
#define REALLOCN(n)              ((n)*1.5)

#define OPTR(v,i,osz)	((char *)(v) + (i)*(osz))

static vector_t *	vector_resize( vector_t *v, int newlen );

/* Allocate a vector_t of length "len".  Anticipating extension of the vector,
 * we actually allocate more memory.
 */
vector_t *vector_create( int obj_size, OBJECT_TYPE obj_type, int length )
{
	vector_header_t	*header;
	int		allocn;

	allocn = ALLOCN(length);

	header = mem_alloc(sizeof(vector_header_t) + allocn*obj_size);
	if (header == NULL) 
	  {
	    return NULL;
	  }

	header->self.object_size = obj_size;
	header->self.object_type = obj_type;
	header->self.length      = length;
	header->self.allocn      = allocn;

	return (vector_t *)(header + 1);
}

/* Routine to free a vector_t */
void vector_destroy( vector_t *v )
{
	if (v == NULL) 
	  {
	    return;
	  }

	mem_free((vector_header_t *)v - 1);
}


/* Return a new vector extended by the given object. */
vector_t *vector_extend( vector_t *v, void *o )
{
	int	newlen, osize;

	newlen = vector_length(v) + 1;
	osize  = vector_object_size(v);

	v = vector_resize(v, newlen);
	if (v == NULL) {
		return NULL;
	}

	(void) memcpy(OPTR(v,newlen-1,osize), o, osize);

	return v;
}

/* Return new vector with last n elements deleted */

vector_t *vector_contract( vector_t *v, int n )
{
	int newlen;

	assert(n >= 0);
	newlen = MAX((vector_length(v) - n), 0);
	return( vector_resize(v, newlen) );
}

/* Internal routine to expand (but not contract) memory for a new vector size */

static vector_t *vector_resize( vector_t *v, int newlen )
{
	vector_header_t	*h;
	int		allocn;


	h = (vector_header_t *)v - 1;

	if (newlen > h->self.allocn)
	  {
	    allocn = REALLOCN(newlen);
	    h = mem_realloc(h, sizeof(vector_header_t) + allocn*h->self.object_size);
	    if (h == NULL) 
	      {
		return NULL;
	      }
	    h->self.allocn = allocn;
	  }

	h->self.length = newlen;

	return (vector_t *)(h + 1);
}

void error_fatal( char *function, char *message, ... )
{
  va_list ap;

  va_start(ap, message);

  fprintf(stderr, "FATAL ERROR in function %s:\n", function);
  fprintf(stderr, "\t");
  vfprintf(stderr, message, ap);
  fprintf(stderr, "\n");

  va_end(ap);

  exit(1);
}

